package de.fraunhofer.iem.dataprovider.taskManager.tasks.kpiCalculation

import de.fraunhofer.iem.dataprovider.dependency.dto.DependencyCreateDto
import de.fraunhofer.iem.dataprovider.dependency.dto.VulnerabilityCreateDto
import de.fraunhofer.iem.dataprovider.dependency.service.DependencyService
import de.fraunhofer.iem.dataprovider.kpi.dto.KPICreateDto
import de.fraunhofer.iem.dataprovider.kpi.service.KPIService
import de.fraunhofer.iem.dataprovider.kpi.strategy.AggregationKPICalculationStrategy
import de.fraunhofer.iem.dataprovider.kpi.strategy.MaximumKPICalculationStrategy
import de.fraunhofer.iem.dataprovider.kpi.strategy.RatioKPICalculationStrategy
import de.fraunhofer.iem.dataprovider.kpi.strategy.RawValueKPICalculationStrategy
import de.fraunhofer.iem.dataprovider.repository.entity.RepositoryDetailsEntity
import de.fraunhofer.iem.dataprovider.repository.service.RepositoryService
import de.fraunhofer.iem.dataprovider.taskManager.events.Event
import de.fraunhofer.iem.dataprovider.taskManager.tasks.Task
import de.fraunhofer.iem.dataprovider.toolRun.service.ToolRunService
import java.util.*

class MetricsTask(
        val repoId: UUID,
        private val repositoryService: RepositoryService,
        private val kpiService: KPIService,
        override val responseChannel: suspend (task: Event) -> Unit,
        private val dependencyService: DependencyService,
        private val toolRunService: ToolRunService,
) : Task() {
    override suspend fun execute() {
        logger.info("Starting metrics task for repoId $repoId")
        val repository = repositoryService.findRepoByID(repoId)
        val repositoryDetails = repositoryService.getLatestRepositoryDetailsByRepositoryId(repoId)
        if (repositoryDetails != null && repository != null) {
            val rootKPI = generateKPITree(repositoryDetails)
            kpiService.calculateKPIs(rootKPI)
            kpiService.storeAndPurgeOld(repository, rootKPI)
        }
    }

    private fun generateKPITree(repositoryDetailsEntity: RepositoryDetailsEntity): KPICreateDto {
        val rootKPI = KPITreeFactory(repositoryDetailsEntity)

        return rootKPI
    }

    private fun KPITreeFactory(repositoryDetailsEntity: RepositoryDetailsEntity): KPICreateDto {
        // lowest level leaves
        val numberOfCommitsKPI = KPICreateDto(
            "Number of Commits",
            "Total number of commits on the default branch of the repository.",
            false,
            RawValueKPICalculationStrategy(repositoryDetailsEntity.numberOfCommits!!)
        )
        val numberOfSignedCommitsKPI = KPICreateDto(
            "Number of Signed Commits",
            "Total number of signed and verified commits on the default branch of the repository.",
            false,
            RawValueKPICalculationStrategy(repositoryDetailsEntity.numberOfSignedCommits!!)
        )
        val isDefaultBranchProtectedKPI = KPICreateDto(
            "Default Branch Protection",
            "Used to assess compliance with a standard development process. For this purpose, it is examined whether the standard development branch is protected against unintentional changes.",
            false,
            RawValueKPICalculationStrategy(if (repositoryDetailsEntity.isDefaultBranchProtected == true) 100 else 0)
        )

        val signedCommitsRatioKPI =
            KPICreateDto(
                "Commit Signature Ratio",
                "Used to assess compliance with a common and transparent development process. It is desirable that all commits are signed by their authors. Therefore, the ratio of signed commits to all commits is determined to calculate this metric.",
                false,
                RatioKPICalculationStrategy()
            )
        signedCommitsRatioKPI.addChildKPI(numberOfCommitsKPI, 1.0)
        signedCommitsRatioKPI.addChildKPI(numberOfSignedCommitsKPI, 1.0)

        // second level
        // TODO: @Jan-Niclas Please add the dependency kpis to the security KPI
        val securityKPI = KPICreateDto(
            "Security Score",
            "Assesses the security of the software provided. For this purpose, various security-relevant analyzes are carried out, which, among other things, check the external dependencies or the code for vulnerabilities.",
            false,
            AggregationKPICalculationStrategy()
        )

        val maximalDependencyVulnerabilityKPI = KPICreateDto(
                "Maximal Dependency Vulnerability Score",
                "This score is calculated by the following formula: 100 - (max(CVSS score) * 10). Thus, a lower value indicates a more critical vulnerability.",
                false,
                // This strategy is tied to this specific KPI. It calculates 100 - max(children). Children are already multiplied by 10
                MaximumKPICalculationStrategy()
        )

        val dependencyDtos = dependencyService.getDependenciesForRepository(repoId);
        // We change the structure to associate vulnerabilities with dependencies instead of dependencies with vulnerabilities
        val vulnerabilitiesWithDependencies = restructureDependenciesAndVulnerabilities(dependencyDtos)
        // For each vulnerability, we create a sub raw data kpi
        addMaximalDependencyVulnerabilitySubKPIs(vulnerabilitiesWithDependencies, maximalDependencyVulnerabilityKPI)
        securityKPI.addChildKPI(maximalDependencyVulnerabilityKPI, 0.5)

        // TODO: There might be the case that there are no latest tool runs for this repository (would end up in an exception)
        val latestToolRunOCCMD = toolRunService.getLatestToolRunsForRepo(repoId).filter { it.tool?.name == "OCCMD" }[0]
        val secretKpi = KPICreateDto(
            "Public Secrets",
            "Score to indicate whether the repository contains secrets like passwords or api keys.",
            false,
            RawValueKPICalculationStrategy(if (latestToolRunOCCMD.findings.isEmpty()) 100 else 0)
        )

        securityKPI.addChildKPI(secretKpi, 0.5)

        val processComplianceKPI = KPICreateDto(
            "Process Compliance Score",
            "Assesses the development process of the software provided. For this purpose, the development process traceable in the repository is compared with common development standards to enable an assessment.",
            false,
            AggregationKPICalculationStrategy()
        )

        val processTransparencyKPI = KPICreateDto(
            "Process Transparency Score",
            "Assesses the transparency resp. traceability of the development process of the provided software for external parties. For this purpose, various analyzes are performed that assess the availability of information about the software development process within the repository.",
            false,
            AggregationKPICalculationStrategy()
        )
        processComplianceKPI.addChildKPI(isDefaultBranchProtectedKPI, 0.5)
        processComplianceKPI.addChildKPI(signedCommitsRatioKPI, 0.5)
        processTransparencyKPI.addChildKPI(signedCommitsRatioKPI, 1.0)

        val rootKPI = KPICreateDto(
            "Project Score",
            "Assesses the project resp. the provided software in the aspects of maturity (based on quality, security and usability aspects) as well as development process.",
            true,
            AggregationKPICalculationStrategy()
        )
        rootKPI.addChildKPI(processTransparencyKPI, 0.25)
        rootKPI.addChildKPI(processComplianceKPI, 0.25)
        rootKPI.addChildKPI(securityKPI, 0.5)
        return rootKPI
    }

    private fun addMaximalDependencyVulnerabilitySubKPIs(vulnerabilitiesWithDependencies: HashMap<VulnerabilityCreateDto, MutableSet<String>>, maximalDependencyVulnerabilityKPI: KPICreateDto) {
        // We iterate through the vulnerabilities and create a sub (raw data) kpi for each of them. Dependencies are displayed in the description
        for ((vulnerability, dependencies) in vulnerabilitiesWithDependencies) {
            maximalDependencyVulnerabilityKPI.addChildKPI(KPICreateDto(
                    "Vulnerability " + vulnerability.cveIdentifier,
                    "Affected packages: " + dependencies.joinToString(", "),
                    false,
                    // TODO: We multiply the CVSS score by 10 as the data structure can only hold integer values. This might be changed in the future.
                    RawValueKPICalculationStrategy((vulnerability.vulnerabilityScores.first().severity.toDouble() * 10).toInt()),
                    // TODO: Displayscore shows the real CVSS value, might not be needed if above TODO is resolved
                    displayScore = vulnerability.vulnerabilityScores.first().severity
            ), 0.0)
        }
    }

    private fun restructureDependenciesAndVulnerabilities(dependencyDtos: List<DependencyCreateDto>): HashMap<VulnerabilityCreateDto, MutableSet<String>> {
        val vulnerabilitiesWithDependencies = HashMap<VulnerabilityCreateDto, MutableSet<String>>()
        // Switch structure around, so we can see which vulnerabilities are there and which dependencies are associated to them
        for (dependencyDto in dependencyDtos) {
            // Iterate through the list of vulnerabilities associated with each dependency
            for (vulnerability in dependencyDto.vulnerabilities) {
                // Check if the vulnerability is already present in the result map
                // If not, add it with an empty list of dependencies
                vulnerabilitiesWithDependencies.computeIfAbsent(vulnerability) { mutableSetOf() }
                // Add the current dependency to the list of dependencies for the vulnerability
                vulnerabilitiesWithDependencies[vulnerability]?.add(dependencyDto.name)
            }
        }
        return vulnerabilitiesWithDependencies
    }
}
