package de.fraunhofer.iem.dataprovider.taskManager.tasks.tools

import de.fraunhofer.iem.dataprovider.taskManager.events.GroupTaskDoneEvent
import de.fraunhofer.iem.dataprovider.taskManager.events.TaskDoneEvent
import de.fraunhofer.iem.dataprovider.taskManager.tasks.Task
import kotlinx.coroutines.CoroutineScope
import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.launch
import java.nio.file.Files
import java.nio.file.Path
import java.nio.file.Paths
import java.util.*

abstract class ToolProcessTask<ResultType> : Task() {
    protected abstract val flags: Array<String>
    protected abstract val resultFileOutputPath: Path
    protected abstract val repoId: UUID
    protected open val execPath: String = "/bin/sh"
    protected abstract val outputDirectory: String

    private val mainScope = CoroutineScope(Dispatchers.Default)
    override suspend fun execute() {
        Files.createDirectories(Paths.get(this.outputDirectory))

        if (isExecutable(execPath)) {
            val process = ProcessBuilder(execPath, *flags).start()
            process.onExit().thenApply { p1 ->
                mainScope.launch {
                    handleProcessReturn(p1)
                }
            }
        } else {
            logger.warn("Given execPath is not an executable $execPath.")
            // TODO: we should probably throw an exception here.
        }
    }

    private fun isExecutable(filePath: String): Boolean {
        val path = Paths.get(filePath)
        return Files.isExecutable(path)
    }

    abstract suspend fun parseProcessResults(resultPath: Path): ResultType
    abstract suspend fun storeResultsInDb(result: ResultType)
    protected open suspend fun handleProcessReturn(p: Process) {
        // TODO: check process exit codes. Current problem, if an analysis
        //  has findings it might return an exit code != 0 even tho the process finished correctly
        logger.info("Handle Process return in $javaClass")
        println(String(p.inputStream.readAllBytes()))
        println(String(p.errorStream.readAllBytes()))
        try {
            val results = parseProcessResults(resultFileOutputPath)
            storeResultsInDb(results)
            sendProcessTaskDoneEvent()
        } catch (e: Throwable) {
            sendTaskFailedEvent(e)
        }
        cleanUp()
    }

    protected open fun cleanUp() {}

    private suspend fun sendProcessTaskDoneEvent() {

        val event = if (groupID != null) {
            GroupTaskDoneEvent(this.taskID, groupID!!)
        } else {
            TaskDoneEvent(this.taskID)
        }

        responseChannel(event)
    }
}
